/*
    $Id: indices.cpp 147 2010-04-17 09:25:25Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <numeric>

#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include "atmos.h"
#include "indices.h"
#include "io.h"
#include "utils.h"

#include "photometry.h"

indices::flux_type indices::flux_nonlinearity_correction ( const flux_type& flux, const vector_type& deadtime ) {
	return photometry::flux_nonlinearity_correction( flux, deadtime );
}
indices::flux_type indices::flux_background_correction( const flux_type& flux, const vector_type& scatter, const vector_type& background ) {
	return photometry::flux_scatter_correction( flux, scatter ) - background;
}
boost::tuple< indices::vars_type, indices::rho1_type, indices::rho2_type >
	indices::vars_correction( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2, const vector_type& deadtime, const vector_type& npoisson ) {
	using namespace utils::ublas;

	typedef matrix_vector_slice< vars_type > diagonal_type;
	typedef boost::tuple< vars_type, rho1_type, rho2_type > result_type;

	assert( vars.size1() == flux.size() );
	assert( vars.size2() == flux.size() );
	assert( rho1.size1() == flux.size() );
	assert( rho1.size2() == flux.size() );
	assert( rho2.size() == flux.size() );
	
	result_type ret(vars,rho1,rho2);

	diagonal_type variances( ret.get<0>(),
		slice(0, 1, std::min( ret.get<0>().size1(), ret.get<0>().size2() )),
		slice(0, 1, std::min( ret.get<0>().size1(), ret.get<0>().size2() )) );
	variances -= photometry::variance_nonpoisson_correction( flux, deadtime, npoisson );
	
	flux_type corr = photometry::variance_nonlinearity_correction( flux, deadtime );

	ret.get<0>() = element_div( ret.get<0>(), outer_prod( corr, corr ) );
	ret.get<1>() = element_div( ret.get<1>(), outer_prod( corr, corr ) );
	ret.get<2>() = element_div( ret.get<2>(), element_prod( corr, corr ) );

	return ret;
}

boost::tuple< indices::indices_type, indices::indesis_type >
	indices::calculate( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
	using namespace utils::ublas;

	typedef matrix_vector_slice< indices_type >    indices_diagonal_type;
	typedef matrix_vector_slice< const rho1_type > rho1_diagonal_type;
	typedef boost::tuple< indices_type, indesis_type> result_type;

	assert( vars.size1() == rho1.size1() );
	assert( vars.size2() == rho1.size2() );
	assert( vars.size1() == flux.size() );
	assert( vars.size2() == flux.size() );

	result_type ret;
	
	rho1_diagonal_type rho1_diag = diag( rho1 );

	ret.get<0>() = element_div( vars*1.25 - rho1*0.25, outer_prod( flux, flux ) );

	indices_diagonal_type variances = diag( ret.get<0>() );

	ret.get<1>() = ( variances*3.0 + element_div( rho2 - rho1_diag * 4.0, element_prod(flux,flux) ) ) / 4.5;

	ret.get<0>() = utils::ublas::log( xp1( ret.get<0>() ) );

	triangle( ret.get<0>(), unit_lower() ) = triangle(
			outer_prod( variances, scalar_vector< value_type >( variances.size(), 1.0 ) )
			+ outer_prod( scalar_vector< value_type >( variances.size(), 1.0 ), variances )
			- triangle( ret.get<0>(), unit_lower() ) * 2.0,
		unit_lower() );

	return ret;
}

indices_flux_threshold_filter::indices_flux_threshold_filter( value_type threshold ):
	m_threshold( threshold ) {
}
indices_flux_threshold_filter::~indices_flux_threshold_filter() {
}
bool indices_flux_threshold_filter::operator() ( const flux_type& flux ) const {
	if( *std::min_element( flux.begin(), flux.end() ) < m_threshold ) {
		throw std::logic_error( (boost::format("Object flux %.2f is under the threshold %.2f") % flux % m_threshold).str() );
		return true;
	}
	return false;
}

const indices::flux_type& indices::flux( const indices::indices_sample& x ) {
	return x.flux;
}
const indices::indices_type& indices::indexes( const indices::indices_sample& x ) {
	return x.indices;
}
const indices::indesis_type& indices::indesis( const indices::indices_sample& x ) {
	return x.indesis;
}

indices_storage::indices_storage( size_type reserve ):
	table_storage< indices::indices_sample >( reserve ),
	m_flux_column( *this, boost::bind(&indices::flux,_1) ),
	m_indices_column( *this, boost::bind(&indices::indexes,_1) ),
	m_indesis_column( *this, boost::bind(&indices::indesis,_1) ) {
}
