/*
    $Id: weif.cpp 158 2010-09-19 08:38:25Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <boost/function.hpp>

#include <boost/numeric/ublas/io.hpp>

#include "weif.h"

#include "gsl_wrap.h"
#include "fft.h"

namespace weif {

const double POWIND = -2.66666666667;
const double POWDIFF = -0.666666666667;
const double MKM2CM = 1E-4;
const double EPS = 1.0e-4 ;
const double CALIBR  = 44.62621363;
const double DIMMCONST = 2.403*4.64159;
const double KM2CM  = 1E5 ;
const int  WORKSPACE = 10000;

struct mass_params {
	typedef boost::function1< std::complex<double>, double > sfun_type;

	double diameter1;
	double diameter2;
	double diameter3;
	double diameter4;
	double PIx2lambda0;
	const sfun_type& sfun;
	double altitude;

	mass_params( const sfun_type& sfun ): sfun(sfun) {}
};

struct dimm_params {
	typedef boost::function1< std::complex<double>, double > sfun_type;

	double aperture;
	double dimmbase;
	double PIx2lambda0;
	double phi;
	const sfun_type& sfun;
	double altitude;

	dimm_params( const sfun_type& sfun ): sfun(sfun) {}
};

double mass_integrand(double x, void * params) {
	mass_params p = *( mass_params *) params;
	double apert1 = 0.0, apert2 = 0.0, apert3 = 0.0, apert4 = 0;
	double f = p.altitude*x*x/2;
	const std::complex<double>& sfun_f = p.sfun(f);
	double sf = imag( sfun_f ) * cos( p.PIx2lambda0*f ) + real( sfun_f ) * sin( p.PIx2lambda0*f );
	sf = sf*sf*pow( x, POWIND );
	if( p.diameter1 != p.diameter2 ) {                // First aperture:
		apert1 = p.diameter1*j1( p.diameter1*x );
		apert2 = p.diameter2*j1( p.diameter2*x );
		apert1 = ( apert1 - apert2 )/( p.diameter1*p.diameter1 - p.diameter2*p.diameter2 );
	}
	if( p.diameter3 != p.diameter4) {                 // Second aperture: the same with indices 1->3, 2->4:
		apert3 = p.diameter3*j1( p.diameter3*x );
		apert4 = ( p.diameter4 == 0 )? 0 : p.diameter4*j1( p.diameter4*x );
		apert3 = ( apert3 - apert4 )/( p.diameter3*p.diameter3 - p.diameter4*p.diameter4 );
	}
	apert1 = 2.0*( apert1 - apert3 )/x;
	return sf*apert1*apert1;
}

double dimm_integrand(double x, void * params) {
	dimm_params p = *( dimm_params *) params;
	double f = p.altitude*x*x/2;
	double a = p.aperture*x;
	double b = p.dimmbase*x;
	const std::complex<double>& sfun_f = p.sfun(f);
	double cf = real( sfun_f ) * cos( p.PIx2lambda0*f ) - imag( sfun_f ) * sin( p.PIx2lambda0*f );
	cf = cf*8*jn(2,a)/a/a;
	cf = cf*cf*pow( x, POWDIFF );
	return cf*( 1.0 - 2.0*cos(p.phi)*cos(p.phi)*j0(b) + 2.0*cos(2*p.phi)*j1(b)/b );
}

template <class T> T nexthigher(T k) {
    if (k == 0) return 1;
    k--;
    for (int i=1; i < std::numeric_limits<T>::digits; i<<=1) k = k | k >> i;
    return k+1;
}
template<class V,class V2> typename V::const_iterator lambda_effective( const V& lambda, const V2& dist ) {
	typedef typename V::value_type value_type;
	typedef typename V::const_iterator const_iterator;

	value_type leff = inner_prod( lambda, dist ) / sum( dist );
	const_iterator it = std::lower_bound( lambda.begin(), lambda.end(), leff );

	return it;
}

std::pair< vector<double>, vector< std::complex<double> > > spectral_band_fourier( const vector<double>& lambda, const vector<double>& dist, unsigned int center ) {
	const int GRIDFACT = 4 ;    // 0 - no stretch, 1 - stretch twice, 2 - stretch 4-times etc.
	double dlambda = lambda(1) - lambda(0);
	int dist_ftnn = 2*GRIDFACT*nexthigher(dist.size());
	vector< std::complex<double> > dist_ft = zero_vector< std::complex<double> >( dist_ftnn );
	vector<double> dist_freq( dist_ftnn/2 );
	for( vector<double>::size_type l = 0; l < dist_freq.size(); l++ ) dist_freq(l) = l/(dlambda*MKM2CM)/dist_ftnn;
	for( vector<double>::size_type l = 0; l < dist.size(); l++ ) {      // Fill in the data array for FFT:
		vector<double>::size_type j = ( l < center ) ? dist_ftnn + l - center : l - center;
		dist_ft(j) = dist(l);
	}
	noalias( dist_ft ) = utils::ublas::fft_radix2( dist_ft ); // Compute the Fourier transform of the center-shifted SED:
	return std::make_pair( dist_freq, subrange( dist_ft, 0, dist_ftnn/2 ) );
}

double mass_weight( double height, double eps1, double eps2, double eps3, double eps4, double lambda0, const boost::function1< std::complex<double>, double >& sfun, double max_freq ) {
	gsl::integration_workspace w(WORKSPACE);
	double error;
	mass_params params(sfun);
	gsl_function F;
	F.function = &mass_integrand;
	F.params = &params;
	params.diameter1 = M_PI*eps1;
	params.diameter2 = M_PI*eps2;
	params.diameter3 = M_PI*eps3;
	params.diameter4 = M_PI*eps4;
	params.PIx2lambda0 = 2*M_PI*lambda0*MKM2CM;
	double result;
	params.altitude = height*KM2CM;
	double hi_limit = std::sqrt( 2 * max_freq / params.altitude );
	if( std::isfinite( hi_limit ) )
		gsl::integration_qags( &F, 0, hi_limit, 0, EPS, WORKSPACE, w, &result, &error);
	else
		gsl::integration_qagiu( &F, 0, 0, EPS, WORKSPACE, w, &result, &error);
	return result*CALIBR;
}

double dimm_weight( double height, double diam, double dbase, double phi, double lambda0, const boost::function1< std::complex<double>, double >& sfun, double max_freq ) {
	gsl::integration_workspace w(WORKSPACE);
	double error;
	dimm_params params(sfun);
	gsl_function F;
	F.function = &dimm_integrand;
	F.params = &params;
	params.aperture = M_PI*diam;
	params.dimmbase = 2*M_PI*dbase;
	params.PIx2lambda0 = 2*M_PI*lambda0*MKM2CM;
	params.phi = phi;
	double result;
	params.altitude = height*KM2CM;
	double hi_limit = std::sqrt( 2 * max_freq / params.altitude );
	if( std::isfinite( hi_limit ) )
		gsl::integration_qags( &F, 0, hi_limit, 0, EPS, WORKSPACE, w, &result, &error);
	else
		gsl::integration_qagiu( &F, 0, 0, EPS, WORKSPACE, w, &result, &error);
	return result*DIMMCONST;
}

}

weif::massdimm_weight::massdimm_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& dimm_lambda, const vector_type& dimm_response, const vector_type& ap_inn, const vector_type& ap_out, double dimmsize, double dimmbase ):
	ap_inn_( ap_inn ), ap_out_( ap_out ), dimmsize_( dimmsize * 100 ), dimmbase_( dimmbase * 100 ),
	mass_leff_index_( lambda_effective( mass_lambda, element_div( mass_response, mass_lambda*MKM2CM ) ).index() ),
	dimm_leff_index_( lambda_effective( dimm_lambda, dimm_response ).index() ),
	mass_leff_( mass_lambda( mass_leff_index_ ) ),
	dimm_leff_( dimm_lambda( dimm_leff_index_ ) ),
	mass_sfourier_( spectral_band_fourier( mass_lambda, element_div( mass_response, mass_lambda*MKM2CM ), mass_leff_index_ ) ),
	dimm_sfourier_( spectral_band_fourier( dimm_lambda, dimm_response, dimm_leff_index_ ) ) {

	mass_freq_max_ = mass_sfourier_.first_data()( mass_sfourier_.size() - 1 );
	dimm_freq_max_ = dimm_sfourier_.first_data()( dimm_sfourier_.size() - 1 );
}

weif::massdimm_weight::vector_type weif::massdimm_weight::operator() ( const value_type& height ) const {
	const size_type mass_nchan = 4;
	vector_type ret( mass_nchan * (mass_nchan + 1) / 2 + 2 );
	size_type cur_n = 0;

	for( size_type i = 0; i < mass_nchan; ++i )
		ret( cur_n++ ) = mass_weight( height, ap_out_(i), ap_inn_(i), 0, 0, mass_leff_, mass_sfourier_, mass_freq_max_ );

	for( size_type i = 0; i < mass_nchan; ++i )
		for( size_type j = i + 1; j < mass_nchan; j++ )
			ret( cur_n++ ) = mass_weight( height, ap_out_(i), ap_inn_(i), ap_out_(j), ap_inn_(j), mass_leff_, mass_sfourier_, mass_freq_max_ );

	ret( cur_n++ ) = dimm_weight( height, dimmsize_, dimmbase_, 0.0,    dimm_leff_, dimm_sfourier_, dimm_freq_max_ ) * 1e+11;
	ret( cur_n++ ) = dimm_weight( height, dimmsize_, dimmbase_, M_PI/2, dimm_leff_, dimm_sfourier_, dimm_freq_max_ ) * 1e+11;

	return ret;
}

linear_generator::linear_generator( result_type start, result_type step ): start_(start), step_(step) {
}
linear_generator::result_type linear_generator::operator()( size_type i ) const {
	return ( start_ + step_ * i );
}
linear_generator::size_type linear_generator::size( result_type max ) const {
	return static_cast<size_type>( std::floor( (max - start_) / step_ ) );
}

lg_generator::lg_generator( result_type start, result_type lstep, result_type gstep ): start_(start), t_(lstep), q_(result_type(1) + gstep / 2) {
	max_ = static_cast< size_type >( std::floor( result_type(2) / gstep - start / lstep ) );
}
lg_generator::result_type lg_generator::operator()( size_type i ) const {
	if( i <= max_ ) return ( start_ + t_ * i );
	return ( start_ + t_ * max_ ) * std::pow( q_, static_cast<int>(i - max_) ); // FIXME: pow(double,int) for <4.2.1
}
lg_generator::size_type lg_generator::size( result_type max ) const {
	result_type gmax = std::floor( std::log( max / ( start_ + t_ * max_ ) ) / std::log( q_ ) );
	if( gmax > result_type(0) ) return static_cast< size_type >( gmax ) + max_ + 1;
	return static_cast< size_type >(std::floor( (max - start_) / t_ ));
}